"use strict";
var express = require('express');

var app = express();

//make Sequelize a singleton
app.set('models', require('./components/models'));
var models = app.get('models');

var port = process.env.PORT || 3000;

var env         = process.env.NODE_ENV || "dev";
// Be sure to have in the config directory the correct server.<env>.js
// Currently :
//   server.dev.js
//   server.qa.js
//   server.prod.js
//   server.test.js
var config      = require("./components/config/server." + env + ".js");

//initialize logging
var winston = require('winston');
var win = require('./components/config/logging')(winston, config);
app.set('winston', win);
var logger = win.loggers.get("main");

app.listen(port, '0.0.0.0');
logger.info("------------> M3-LOADER START ! <------------------");
logger.info("Pid : " +process.pid + " - Fresenius Loader server started on port " + port + " on " + env);
logger.info("Rabbit Topology : ");
logger.info("       host -> " + config.rabbitConfig.host);
logger.info("       user -> " + config.rabbitConfig.user);
logger.info("   password -> " + config.rabbitConfig.pass);
logger.info("       port -> " + config.rabbitConfig.port);
logger.info("      vhost -> " + config.rabbitConfig.vhost);

//initialize connections to the MQ framework
if (config.mqs && config.mqs === "rabbit") {
    var Rabbit = require("./components/helpers/RabbitHelper");
    app.set("mqs", new Rabbit(winston, config));
} else {
    var SquissHelper = require("./components/helpers/SqsHelper");
    app.set("mqs", new SquissHelper(winston, config));
}

//initialize express
require('./components/config/express')(app, config);

process.on('uncaughtException', function (err) {
    var msg = (new Date).toUTCString() + ' uncaughtException:';
    logger.error(msg, err.message);
    logger.error(err.stack);
    process.exit(1);
});

module.exports = app;
