/**
 * Created by TFG on 07/15/2020.
 * Management for the Master Item loading data
 *
 * CAUTION ! August 2020 note !
 * Since Master Item process is not complete, in order to have an intermediate release MT will ignore any record
 * coming in the queue !
 * See function processMetadata - set abort flag to true !
 * That flag should be set to false for normal working !
 */
var app = require("../../child");

var LoaderHelper = require("../helpers/LoaderHelper");
var awcblcommon = require('../config/awcblcommon');

var currentTransaction;
var itemCatalogs = [];    // List of items catalog records associated to the input record
// Doing that because unable to assign correctly the models variable where needed
var itemCatalog = {
    itemId: "",
    catalogId: ""
};

// Object to deal with metadata
var MTmetadata = {
    transactionId: "N/A",     //    - transaction ID
    fileName: "--",           //    - name of the originating CSV file
    expectedNumRec: 0,        //    - number to record to be received (remove 1 when counting)
    rcvCounter: 0,            //    - counter of record received  (remember to add 1 when showing the value - counter from 0 not 1)
    dbCounter: 0,             //    - counter of record saved on DB ((remember to add 1 when showing the value - counter from 0 not 1)
    endLoad: false,           //    - flag to indicate the current load as ended
    endProc: false,           //    - flag to indicate the current processing as ended
    abort: false,             //    - flag to abort the transaction ! If true all the incoming records will ack but not processed
    sendEndLoad: true,        //    - flag to be used to send out only ONE log notification about end loading
    sendEndProc: true,        //    - flag to be used to send out only ONE log notification about end processing
    sendAbort: true,          //    - flag to be used to send out only ONE log notification about abort
    incrementRcvCounter : function() {
        this.rcvCounter++;
        if (this.rcvCounter >= (this.expectedNumRec -1)) {
            this.endLoad = true;
        }
    },
    incrementDbCounter : function() {
        this.dbCounter++;
        // Here the test is little bit different because the position of the incrementDbCounter
        // Starting from zero, when process the number 9 (the latest) it will be incremented by one and
        // then the test
        if (this.dbCounter >= this.expectedNumRec) {
            this.endProc = true;
        }
    }
};

// Create array of metadata objects
var metadata = [ MTmetadata ];

module.exports = function(winston, rabbit, models, config, cb) {
    // variable to hold starting time
    var started = Date.now();

    // expected message count
    var logger = winston.loggers.get("main");

    // always setup your message handlers first
    rabbit.nackOnError();

    // this handler will handle messages sent from the publisher
    rabbit.handle({
        queue: 'topic-fresenius-loader-mtq',
        type: 'loader.mt'
    }, function (msg) {
        var helper = new LoaderHelper(models, winston);

        logger.info(process.pid + ":LOADER MT POST Received:",
            JSON.stringify(msg.body,null, 4),
            "routingKey:", msg.fields.routingKey);

        // Metadata management - use the currentTransaction to monitor the load
        currentTransaction = processMetadata(winston, msg.body);

        // Intercept wrong incoming records !
        if (currentTransaction === null ||
            currentTransaction === undefined) {
            // Something wrong handling metadata !!!
            // Ignore incoming record !
            msg.ack();      // Acknowledge the message in the queue to clean it up
        } else {
            if (currentTransaction.abort === true) {
                if (currentTransaction.sendAbort === true) {
                    // Write log to notify successfull end of operation
                    awcblcommon.writeAwcblEntryLog(app,
                        currentTransaction.fileName,
                        currentTransaction.transactionId,
                        "BLOCKED",
                        "Transaction ignored - metadata error").then(response => {
//                        logger.debug("AWCBL log written : " + response);
                    }, reject => {
                        logger.error("AWCBL log : " + reject);
                    });
                    currentTransaction.sendAbort = false;     // Disable sending other LOG messages about abort
                }
                msg.ack();      // Acknowledge the message in the queue to clean it up
            } else {
                // NOT ABORT !! Process the incoming record !
                // Store the record on the DB via helper
                helper.saveLoaderMT(msg.body).then(response => {
                    // Response - last record saved
                    logger.info("Record " + currentTransaction.dbCounter + " saved : " + response);
                    currentTransaction.incrementDbCounter();        // Increment saved record counter

                    // Extract and prepare the associated datat to the record to be saved
                    // The data to be loaded are stored in the array itemCatalogs
                    processAssociatedData(winston, msg.body, response);

                    // Remove old entries from tbl_item_catalog for the current item
                    helper.removeMTassociated(response).then(response => {
                        // Removed associated data !!!
                        // Now save the new associated data

                        for (var fn = 0; fn < itemCatalogs.length; fn++) {
                            // Call here the function to save the associated data
                            helper.saveMTassociatedEntry(itemCatalogs[fn].itemId, itemCatalogs[fn].catalogId).then(response => {
                                // Saved associated data !
                                logger.info("Associated data to record " + currentTransaction.dbCounter + " saved : " + response);
                            }, reject => {
                                logger.error("Error from helper for MT associated : " + reject);

                                // Write log to notify error happened
                                awcblcommon.writeAwcblEntryLog(app,
                                    currentTransaction.fileName,
                                    currentTransaction.transactionId,
                                    "ERROR",
                                    "Error updating associated : " + reject + " - Contact assistance").then(response => {
//                                    logger.debug("AWCBL log written : " + response);
                                }, reject => {
                                    logger.error("AWCBL log : " + reject);
                                });
                            });
                        }
                        msg.ack();      // Acknowledge the message in the queue to clean it up
                    }, reject => {
                        logger.error("Error removing for MT associated : " + reject);

                        // Write log to notify error happened
                        awcblcommon.writeAwcblEntryLog(app,
                            currentTransaction.fileName,
                            currentTransaction.transactionId,
                            "ERROR",
                            "Error removing old associated data : " + reject + " - Contact assistance").then(response => {
//                            logger.debug("AWCBL log written : " + response);
                        }, reject => {
                            logger.error("AWCBL log : " + reject);
                        });
                        msg.ack();      // Acknowledge the message in the queue to clean it up
                    });

                    if (currentTransaction.endProc === true && currentTransaction.sendEndProc === true) {
                        // Write log to notify successfull end of operation
                        awcblcommon.writeAwcblEntryLog(app,
                            currentTransaction.fileName,
                            currentTransaction.transactionId,
                            "PROCESSED",
                            "Processed " + currentTransaction.dbCounter + "  records").then(response => {
//                            logger.debug("AWCBL log written : " + response);
                        }, reject => {
                            logger.error("AWCBL log : " + reject);
                        });
                        currentTransaction.sendEndProc = false;     // Disable sending other LOG messages about ending process
                    }
                    msg.ack();      // Acknowledge the message in the queue to clean it up
                }, reject => {
                    // Record not saved ! Ack message anyway to remove from the queue
                    logger.error("MT Error from helper : " + reject);

                    // Write log to notify error happened
                    awcblcommon.writeAwcblEntryLog(app,
                        currentTransaction.fileName,
                        currentTransaction.transactionId,
                        "ERROR",
                        "Error loading : " + reject + " - Contact assistance").then(response => {
//                        logger.debug("AWCBL log written : " + response);
                    }, reject => {
                        logger.error("AWCBL log : " + reject);
                    });
                    msg.ack();      // Acknowledge the message in the queue to clean it up
                });

                // Shows in the log statistics about processing
                logger.info("Transaction " + currentTransaction.transactionId +
                    " - expected " + currentTransaction.expectedNumRec +
                    " - received " + currentTransaction.rcvCounter +
                    " - stored " + currentTransaction.dbCounter +
                    " - end load : " + currentTransaction.endLoad.toString() +
                    " - end proc : " + currentTransaction.endProc.toString());

                if (currentTransaction.endLoad === true && currentTransaction.sendEndLoad === true) {
                    logger.info("Transaction " + currentTransaction.transactionId + " ENDED !");

                    // Write log to notify successfull end of operation
                    awcblcommon.writeAwcblEntryLog(app,
                        currentTransaction.fileName,
                        currentTransaction.transactionId,
                        "LOADED",
                        "Loaded " + (currentTransaction.rcvCounter + 1) + " records").then(response => {
//                        logger.debug("AWCBL log written : " + response);
                    }, reject => {
                        logger.error("AWCBL log : " + reject);
                    });
                    currentTransaction.sendEndLoad = false;     // Disable sending other LOG messages about ending loading
                }
                msg.ack();      // Acknowledge the message in the queue to clean it up
            }
        }
    });


    logger.info(process.pid + ':Set up loader mtq OK');
    return require("../config/RabbitTopology.js")(rabbit, "loader.mt", config);
};

// Function to handle the metadata related to a load.
function processMetadata(winston, msg) {
    var path = require('path');
    var logger = winston.loggers.get("main");
    var inTransactionId;
    var inFileName;
    var inExpectedNumRec;

    // Extract metadata from input message - it is expected :
    //   csv_line[0] -> transactionId
    //   csv_line[1] -> file name
    //   csv_line[2] -> expected number of messages
    try {
        var csv_line = msg.split(/,(?=(?:(?:[^"]*"){2})*[^"]*$)/);
        csv_line = csv_line || [];

        // Retrieve metadata
        inTransactionId = csv_line[0].split(',').join('');
        inFileName = path.basename(csv_line[1].split(',').join(''), ".csv") + ".csv";
        inExpectedNumRec = parseInt(csv_line[2].split(',').join(''), 10);
    } catch (err) {
        logger.error("MT - Process Metadata ! Error ! : " + err);
        // Probably the format of the input is wrong !
        return null;
    }

    // Check if the transaction ID exists on the array
    let currentTransaction = metadata.find(MTmetadata => MTmetadata.transactionId === inTransactionId);

    if (currentTransaction !== undefined) {
        currentTransaction.incrementRcvCounter();
    } else {
        logger.info("New transaction identified : " + inTransactionId);
        // Not found object - add one
        MTmetadata.transactionId = inTransactionId;
        MTmetadata.fileName = inFileName;
        MTmetadata.expectedNumRec = inExpectedNumRec;
        MTmetadata.rcvCounter = 0;
        MTmetadata.dbCounter = 0;
        MTmetadata.endLoad = false;
        MTmetadata.endProc = false;
        MTmetadata.sendEndLoad = true;
        MTmetadata.sendEndProc = true;
        MTmetadata.sendAbort = true;
        MTmetadata.abort = false;   // TFG !!! Force true to ignore incoming records for MT !!!

        metadata.push(MTmetadata);
        currentTransaction = metadata.find(MTmetadata => MTmetadata.transactionId === inTransactionId);
    }

    return currentTransaction;
}

// Function to handle the MT associated data to store after saving all the data
// Store the data to be sent in the global variable itemCatalogs.
function processAssociatedData( winston, msg, inItemIdId) {
    var path = require('path');
    var logger = winston.loggers.get("main");

    var startingCSVindex = 3;
    var inItemId;
    var inTestCatalog;

    // Clean up storage
    itemCatalogs.length = 0;

    try {
        var csv_line = msg.split(/,(?=(?:(?:[^"]*"){2})*[^"]*$)/);
        csv_line = csv_line || [];

        // Retrieve catalog related data !!
        //inItemId = csv_line[startingCSVindex].split(',').join('');
        inTestCatalog = csv_line[startingCSVindex+4].split(',').join('');
    } catch (err) {
        logger.error("Process Associated data ! Error ! : " + err);
        // Probably the format of the input is wrong !
        return;
    }

//    logger.debug("processAssociatedData - item ID    = " + inItemIdId);
//    logger.debug("                      - catalog ID   = " + inTestCatalog);

    // Prepare the list of the record for tbl_item_catalog to add
    var catIds = inTestCatalog.split('\#');
    for (var fn = 0; fn < catIds.length; fn++) {
        itemCatalogs[fn] = new createItemCatalogEntry(inItemIdId, catIds[fn]);
    }
}

function createItemCatalogEntry(item_id, catalog_id) {
    this.itemId = item_id;
    this.catalogId = catalog_id;
}
